/*
 * $Id: gamemanager.c,v 1.4 2001/12/20 15:01:36 kenta Exp $
 *
 * Copyright 2001 Kenta Cho. All rights reserved.
 */

/**
 * Re: W32/Badtrance game lifecycle manager.
 *
 * @version $Revision: 1.4 $
 */
#include <string.h>
#include <stdlib.h>
#include <piece.h>

#include "re.h"
#include "gamemanager.h"
#include "render.h"

#define NOT_EXIST -999999

#define RE_MAX 8

static Re re[RE_MAX];
static Re ol;

#define BULLET_MAX 16

static Bullet bullet[BULLET_MAX];

#define LOCK_MAX 16

static Lock lock[LOCK_MAX];

static Player player;
Vector plyVt[9];

static float groundZ;
static int scrGndIdx;

#define GROUND_NUM 6

Vector gndVt[GROUND_NUM*4];

#define GROUND_TEX_WIDTH (128<<8)
#define GROUND_TEX_HEIGHT (24<<8)

#define GROUND_HEIGHT (128<<8)
#define GROUND_OFFSET (52<<8)

static int gndPtn[6][GROUND_NUM*4][2];

int stage;

static int accel = 10;
static int bulletSpeed, bulletMxMax;
static int bulletShotFrq;
static int addReFrq;
int stageReNum;
static int zSpeed, crtZspd;

int score;

/*int hiScore;

#define HISCORE_FILE "re.cfg"

void loadHiScore() {
  FILEACC pfa;
  if ( pceFileOpen(&pfa, HISCORE_FILE, FOMD_RD) == 0 ) {
    pceFileReadSct(&pfa, &hiScore, 0, sizeof(int));
    pceFileClose(&pfa);
  } else {
    pceFileCreate(HISCORE_FILE, sizeof(int));
    hiScore = 5000;
  }
}

void saveHiScore() {
  FILEACC pfa;
  if ( pceFileOpen(&pfa, HISCORE_FILE, FOMD_WR) == 0 ) {
    pceFileWriteSct(&pfa, &hiScore, 0, sizeof(int));
    pceFileClose(&pfa);
  }
}*/

void drawObject(PIECE_BMP *bmp, int dx, int dy, int w, int h) {
  DRAW_OBJECT obj;
  pceLCDSetObject(&obj, bmp, dx, dy, 0, 0, w, h, DRW_NOMAL);
  pceLCDDrawObject(obj);
}

static void setStage() {
  int ss = stage % 5;
  accel = 10 - 1.2f*ss;
  bulletSpeed = 256 + ss*72 + stage*16;
  if ( bulletSpeed > 768 ) bulletSpeed = 768;
  bulletMxMax = 80 + ss*50 + stage*9;
  if ( bulletSpeed > 512 ) bulletSpeed = 512;
  bulletShotFrq = 88 - ss*3 - stage;
  if ( bulletShotFrq < 32 ) bulletShotFrq = 32;
  addReFrq = 32 - ss*4- stage*2;
  if ( addReFrq < 8 ) addReFrq = 8;
  stageReNum = 24 + ss*7 + stage*4;
  if ( stageReNum > 90 ) stageReNum = 90;
  zSpeed = (ss+1)<<8;
  if ( stage > 0 ) crtZspd = 32<<8;
  else crtZspd = zSpeed;

  setGroundVertices(ss, 0, 256);
}

void gotoNextStage() {
  stage++;
  score += stage * 500;
  setStage();
}

static void makeGroundPattern() {
  int i, idx, x, y, w;
  int yPtn[] = {0, 1, 2, 3, 4, 3, 2, 1, 0};
  idx = 0;
  for ( i=0 ; i<GROUND_NUM ; i++ ) {
    gndPtn[0][idx][0] = -GROUND_TEX_WIDTH;
    gndPtn[0][idx][1] = GROUND_HEIGHT; idx++;
    gndPtn[0][idx][0] = GROUND_TEX_WIDTH;
    gndPtn[0][idx][1] = GROUND_HEIGHT; idx++;
    gndPtn[0][idx][0] = GROUND_TEX_WIDTH;
    gndPtn[0][idx][1] = GROUND_HEIGHT; idx++;
    gndPtn[0][idx][0] = -GROUND_TEX_WIDTH;
    gndPtn[0][idx][1] = GROUND_HEIGHT; idx++;
  }
  idx = 0;
  for ( i=0 ; i<GROUND_NUM ; i+=2 ) {
    gndPtn[1][idx][0] = -GROUND_TEX_WIDTH;
    gndPtn[1][idx][1] = GROUND_HEIGHT>>1; idx++;
    gndPtn[1][idx][0] = -GROUND_TEX_WIDTH;
    gndPtn[1][idx][1] = GROUND_HEIGHT; idx++;
    gndPtn[1][idx][0] = -GROUND_TEX_WIDTH;
    gndPtn[1][idx][1] = GROUND_HEIGHT; idx++;
    gndPtn[1][idx][0] = -GROUND_TEX_WIDTH;
    gndPtn[1][idx][1] = GROUND_HEIGHT>>1; idx++;
    gndPtn[1][idx][0] = GROUND_TEX_WIDTH;
    gndPtn[1][idx][1] = GROUND_HEIGHT; idx++;
    gndPtn[1][idx][0] = GROUND_TEX_WIDTH;
    gndPtn[1][idx][1] = GROUND_HEIGHT>>1; idx++;
    gndPtn[1][idx][0] = GROUND_TEX_WIDTH;
    gndPtn[1][idx][1] = GROUND_HEIGHT>>1; idx++;
    gndPtn[1][idx][0] = GROUND_TEX_WIDTH;
    gndPtn[1][idx][1] = GROUND_HEIGHT; idx++;
  }
  idx = 0;
  for ( i=0 ; i<GROUND_NUM ; i++ ) {
    y = randN(GROUND_HEIGHT) + GROUND_HEIGHT;
    x = randNS(GROUND_TEX_WIDTH);
    w = randN(GROUND_TEX_WIDTH>>1) + (GROUND_TEX_WIDTH>>2);
    gndPtn[2][idx][0] = x-w;
    gndPtn[2][idx][1] = y; idx++;
    gndPtn[2][idx][0] = x+w;
    gndPtn[2][idx][1] = y; idx++;
    gndPtn[2][idx][0] = x+w;
    gndPtn[2][idx][1] = y; idx++;
    gndPtn[2][idx][0] = x-w;
    gndPtn[2][idx][1] = y; idx++;
  }  
  idx = 0;
  for ( i=0 ; i<GROUND_NUM ; i+=2 ) {
    gndPtn[3][idx][0] = -GROUND_TEX_WIDTH;
    gndPtn[3][idx][1] = (GROUND_HEIGHT>>1); idx++;
    gndPtn[3][idx][0] = -(GROUND_TEX_WIDTH>>1);
    gndPtn[3][idx][1] = GROUND_HEIGHT; idx++;
    gndPtn[3][idx][0] = -(GROUND_TEX_WIDTH>>1);
    gndPtn[3][idx][1] = GROUND_HEIGHT; idx++;
    gndPtn[3][idx][0] = -GROUND_TEX_WIDTH;
    gndPtn[3][idx][1] = (GROUND_HEIGHT>>1); idx++;
    gndPtn[3][idx][0] = GROUND_TEX_WIDTH;
    gndPtn[3][idx][1] = (GROUND_HEIGHT>>1); idx++;
    gndPtn[3][idx][0] = (GROUND_TEX_WIDTH>>1);
    gndPtn[3][idx][1] = GROUND_HEIGHT; idx++;
    gndPtn[3][idx][0] = (GROUND_TEX_WIDTH>>1);
    gndPtn[3][idx][1] = GROUND_HEIGHT; idx++;
    gndPtn[3][idx][0] = GROUND_TEX_WIDTH;
    gndPtn[3][idx][1] = (GROUND_HEIGHT>>1); idx++;
  }
  idx = 0;
  for ( i=0 ; i<GROUND_NUM ; i+=2 ) {
    gndPtn[4][idx][0] = GROUND_TEX_WIDTH;
    gndPtn[4][idx][1] = GROUND_HEIGHT+yPtn[i]*(GROUND_HEIGHT>>4); idx++;
    gndPtn[4][idx][0] = (GROUND_TEX_WIDTH>>1);
    gndPtn[4][idx][1] = GROUND_HEIGHT+yPtn[i]*(GROUND_HEIGHT>>4); idx++;
    gndPtn[4][idx][0] = (GROUND_TEX_WIDTH>>1);
    gndPtn[4][idx][1] = GROUND_HEIGHT+yPtn[i+1]*(GROUND_HEIGHT>>4); idx++;
    gndPtn[4][idx][0] = GROUND_TEX_WIDTH;
    gndPtn[4][idx][1] = GROUND_HEIGHT+yPtn[i+1]*(GROUND_HEIGHT>>4); idx++;
    gndPtn[4][idx][0] = -(GROUND_TEX_WIDTH>>1);
    gndPtn[4][idx][1] = GROUND_HEIGHT+yPtn[i]*(GROUND_HEIGHT>>4); idx++;
    gndPtn[4][idx][0] = -GROUND_TEX_WIDTH;
    gndPtn[4][idx][1] = GROUND_HEIGHT+yPtn[i]*(GROUND_HEIGHT>>4); idx++;
    gndPtn[4][idx][0] = -GROUND_TEX_WIDTH;
    gndPtn[4][idx][1] = GROUND_HEIGHT+yPtn[i+1]*(GROUND_HEIGHT>>4); idx++;
    gndPtn[4][idx][0] = -(GROUND_TEX_WIDTH>>1);
    gndPtn[4][idx][1] = GROUND_HEIGHT+yPtn[i+1]*(GROUND_HEIGHT>>4); idx++;
  }
  idx = 0;
  for ( i=0 ; i<GROUND_NUM ; i++ ) {
    gndPtn[5][idx][0] = -GROUND_TEX_WIDTH;
    gndPtn[5][idx][1] = -GROUND_HEIGHT; idx++;
    gndPtn[5][idx][0] = GROUND_TEX_WIDTH;
    gndPtn[5][idx][1] = -GROUND_HEIGHT; idx++;
    gndPtn[5][idx][0] = GROUND_TEX_WIDTH;
    gndPtn[5][idx][1] = -GROUND_HEIGHT; idx++;
    gndPtn[5][idx][0] = -GROUND_TEX_WIDTH;
    gndPtn[5][idx][1] = -GROUND_HEIGHT; idx++;
  }
}

void initGroundVertices() {
  int i, idx = 0;
  for ( i=0 ; i<GROUND_NUM ; i++ ) {
    gndVt[idx].z = (i+1)*GROUND_OFFSET; idx++;
    gndVt[idx].z = (i+1)*GROUND_OFFSET; idx++;
    gndVt[idx].z = (i+1)*GROUND_OFFSET + GROUND_TEX_HEIGHT; idx++;
    gndVt[idx].z = (i+1)*GROUND_OFFSET + GROUND_TEX_HEIGHT; idx++;
  }
  groundZ = 0;
  scrGndIdx = 0;
}

void setGroundVertices(int r1, int r2, int pct) {
  int i, idx = 0;
  for ( i=0 ; i<GROUND_NUM*4 ; i++, idx++ ) {
    gndVt[idx].x = (gndPtn[r1][idx][0]*pct + gndPtn[r2][idx][0]*(256-pct))>>8;
    gndVt[idx].y = (gndPtn[r1][idx][1]*pct + gndPtn[r2][idx][1]*(256-pct))>>8;
  }
}

void initGamemanager() {
  makeGroundPattern();
  initGroundVertices();
}

static int demoPad;

void initGame(int s) {
  int i;
  for ( i=0 ; i<RE_MAX ; i++ ) {
    re[i].x = NOT_EXIST;
  }
  ol.ax = 0; ol.ay = 16<<8;
  ol.x = ol.ax + (8<<8); ol.y = ol.ay + (4<<8);
  ol.mx = -randN(64)-64; ol.my = randN(32)+32;
  for ( i=0 ; i<BULLET_MAX ; i++ ) {
    bullet[i].x = NOT_EXIST;
  }
  player.x = (SCREEN_WIDTH/2)<<8; player.y = 80<<8;
  player.tx = SCREEN_WIDTH/2; player.ty = SCREEN_HEIGHT/2;
  player.hp = 3;
  player.cnt = player.trgIdx = player.btnPressed = 0;
  for ( i=0 ; i<9 ; i+=3 ) {
    plyVt[i].x = plyVt[i].y = plyVt[i].z = 0;
  }
  for ( i=0 ; i<LOCK_MAX ; i++ ) {
    lock[i].spc = NOT_EXIST;
  }
  stage = s;
  setStage();
  score = 0;
  demoPad = PAD_RI;
}

static int reIdx = RE_MAX;

static void addRe() {
  int i;
  Re *r;
  if ( stageReNum <= 0 ) return;
  for ( i=0 ; i<RE_MAX ; i++ ) {
    reIdx--; if ( reIdx < 0 ) reIdx = RE_MAX-1;
    if ( re[reIdx].x == NOT_EXIST ) break;
  }
  if ( i == RE_MAX ) return;
  r = &(re[reIdx]);
  r->x = ol.x; r->y = ol.y;
  if ( randN(2) == 0 ) {
    r->ax = ((randN(240)+64)/accel)<<8;
  } else {
    r->ax = ((-randN(240)-64)/accel)<<8;
  }
  r->ay = ((randN(160)+64)/accel)<<8;
  r->mx = randNS(10240)/accel; r->my = randN(2560)/accel;
  r->locked = 0;
  stageReNum--;
}

static int bltIdx = BULLET_MAX;

static void addBullet(int x, int y) {
  int i;
  Bullet *blt;
  for ( i=0 ; i<BULLET_MAX ; i++ ) {
    bltIdx--; if ( bltIdx < 0 ) bltIdx = BULLET_MAX-1;
    if ( bullet[bltIdx].x == NOT_EXIST ) break;
  }
  if ( i == BULLET_MAX ) return;
  blt = &(bullet[bltIdx]);
  blt->x = (x<<8); blt->y = (y<<8);
  blt->my = bulletSpeed;
  blt->mx = (player.x - blt->x)/((player.y - blt->y)/blt->my);
  if ( blt->mx < -bulletMxMax ) blt->mx = -bulletMxMax;
  if ( blt->mx >  bulletMxMax ) blt->mx =  bulletMxMax;
  blt->locked = 0;
}

#define LOCK_WIDTH 10
#define OL_WIDTH 12

#define PLAYER_LOCK_MAX 8

static int lckIdx = LOCK_MAX;

static int addLock(Lock *trg) {
  if ( !player.btnPressed || player.trgIdx >= PLAYER_LOCK_MAX ) return 0;
  playSe(KICK2, 1);
  lckIdx--; if ( lckIdx < 0 ) lckIdx = LOCK_MAX-1;
  lock[lckIdx] = *trg;
  lock[lckIdx].laser.cnt = NOT_EXIST;
  lock[lckIdx].num = player.trgIdx;
  player.trgIdx++;
  return 1;
}

static void addLaser(int x, int y) {
  int i;
  Laser *ls;
  for ( i=0 ; i<LOCK_MAX ; i++ ) {
    if ( lock[i].spc == NOT_EXIST || lock[i].laser.cnt != NOT_EXIST ) continue;
    ls = &(lock[i].laser);
    ls->cnt = 0;
    ls->x[0] = x; ls->y[0] = y;
    ls->d[0] = randN(2);
  }
}

int drawRe() {
  int i, x, y;
  Re *r;
  int reExistNum = 0;
  for ( i=0 ; i<RE_MAX ; i++ ) {
    if ( re[i].x == NOT_EXIST ) continue;
    reExistNum++;
    r = &(re[i]);
    r->mx += (r->ax - r->x)>>accel;
    r->my += (r->ay - r->y)>>accel;
    r->x += r->mx; r->y += r->my;
    r->cnt++;
    y = r->sy = r->y>>8;
    x = r->sx = r->x / ((256+(SCREEN_WIDTH<<1)) - (y<<1)) + SCREEN_WIDTH/2;
    if ( status != DEAD ) {
      if ( (r->cnt%bulletShotFrq)==0 && 
	   x>=0 && x<SCREEN_WIDTH && y>=0 && y<SCREEN_HEIGHT/2 ) {
	addBullet(x, y);
      }
    } else {
      r->ay -= 256;
    }
    if ( y < 22 ) {
      drawObject(&re1Bmp, x-8, y-3, 16, 6);
    } else if ( y < 44 ) {
      drawObject(&re2Bmp, x-8, y-4, 16, 9);
    } else if ( y < 66 ) {
      drawObject(&re3Bmp, x-12, y-5, 24, 10);
    } else {
      drawObject(&re4Bmp, x-16, y-7, 32, 14);
    }
    if ( !r->locked && absN(x-player.tx) < LOCK_WIDTH && absN(y-player.ty) < LOCK_WIDTH &&
	 (absN(x-ol.sx) >= OL_WIDTH || absN(y-ol.sy) >= OL_WIDTH) ) {
      Lock trg;
      trg.spc = TARGET_RE; trg.re = r;
      r->locked = addLock(&trg);
    }
  }
  ol.mx += (ol.ax - ol.x)>>(accel+1);
  ol.my += (ol.ay - ol.y)>>(accel+1);
  if ( ol.y < 0 || ol.y > (SCREEN_HEIGHT<<7) || 
       ol.x < -(SCREEN_WIDTH<<6) || ol.x > (SCREEN_WIDTH<<6) ) {
    ol.mx -= (ol.mx>>6); ol.my -= (ol.my>>6);
  }
  ol.x += ol.mx; ol.y += ol.my;
  y = ol.sy = ol.y>>8;
  x = ol.sx = ol.x / ((256+(SCREEN_WIDTH<<1)) - (y<<1)) + SCREEN_WIDTH/2;
  drawObject(&olBmp, x-16, y-16, 32, 32);
  if ( status != DEAD ) {
    if ( (cnt%addReFrq) == 0 ) {
      addRe();
    }
  } else {
    ol.ay -= 256;
  }
  return reExistNum;
}

#define BULLET_LENGTH 8
#define BULLET_COLOR 3

void drawBullet() {
  int i, x, y, sy, ey;
  unsigned char *vb;
  Bullet *bl;
  int px = (player.x>>8), py = (player.y>>8);
  for ( i=0 ; i<BULLET_MAX ; i++ ) {
    if ( bullet[i].x == NOT_EXIST ) continue;
    bl = &(bullet[i]);
    if ( status != DEAD ) {
      bl->x += bl->mx;
      bl->y += bl->my;
      if ( bl->y >= (SCREEN_HEIGHT<<8) ) {
	bl->x = NOT_EXIST;
	continue;
      }
    }
    x = bl->sx = (bl->x>>8);
    if ( x < 1 || x >= SCREEN_WIDTH-1 ) continue; 
    bl->sy = (bl->y>>8);
    sy = bl->sy - BULLET_LENGTH/2;
    ey = sy + BULLET_LENGTH;
    if ( sy < 0 ) sy = 0;
    if ( ey > SCREEN_HEIGHT ) ey = SCREEN_HEIGHT;
    //vb = &(vbuf[x-1+sy*SCREEN_WIDTH]);
    vb = &(vbuf[x-1+(sy<<7)]);
    for ( y=sy ; y<ey ; y++, vb+=SCREEN_WIDTH-2 ) {
      *vb = BULLET_COLOR; vb++;
      *vb = BULLET_COLOR; vb++;
      *vb = BULLET_COLOR;
    }
    if ( status != DEAD ) {
      if ( !bl->locked && absN(x-player.tx) < LOCK_WIDTH && absN(bl->sy-player.ty) < LOCK_WIDTH ) {
	Lock trg;
	trg.spc = TARGET_BULLET; trg.bullet = bl;
	bl->locked = addLock(&trg);
      }
      if ( absN(x-px) <3 && absN(y-py) < 1 ) {
	initPlayerDead();
      }
    }
  }
}

static int prvTrgIdx = 0;

#define PLAYER_MV 768
#define TARGET_MV 6

void drawPlayer() {
  int i;
  int wdCnt = cnt % 16;
  int pad = pcePadGet();
  if ( demo ) { 
    pad = demoPad;
    if ( randN(8) == 0 ) {
      if ( demoPad == PAD_RI && player.tx > (SCREEN_WIDTH>>1) ) demoPad = PAD_UP;
      else if ( demoPad == PAD_UP && player.ty < (SCREEN_HEIGHT>>2) ) demoPad = PAD_LF;
      else if ( demoPad == PAD_LF && player.tx < (SCREEN_WIDTH>>1) ) demoPad = PAD_DN;
      else if ( demoPad == PAD_DN && player.ty > (SCREEN_HEIGHT>>2) ) demoPad = PAD_RI;
    } else {
      pad |= PAD_A;
    }
  } else {
    pad = pcePadGet();
  }
  if ( status == DEAD ) {
    if ( wdCnt < 8 ) {
      drawObject(&plBmp, (player.x>>8)-4, (player.y>>8)-4,  8, 8);
      drawPrimitive(plyVt, 9, 3, 
		    player.x-(SCREEN_WIDTH<<7), player.y-(SCREEN_HEIGHT<<5), (SCREEN_WIDTH<<7));
    }
    return;
  }
  if ( pad & PAD_LF ) {
    player.x -= PLAYER_MV;
    if ( player.x < 1024 ) player.x = 1024;
    player.tx -= TARGET_MV;
    if ( player.tx < 8 ) player.tx = 8;
  }
  if ( pad & PAD_RI ) {
    player.x += PLAYER_MV;
    if ( player.x > (SCREEN_WIDTH<<8)-1024 ) player.x = (SCREEN_WIDTH<<8)-1024;
    player.tx += TARGET_MV;
    if ( player.tx >= SCREEN_WIDTH-8 ) player.tx = SCREEN_WIDTH-8;
  }
  if ( pad & PAD_UP ) {
    player.ty -= TARGET_MV;
    if ( player.ty < 8 ) player.ty = 8;
  }
  if ( pad & PAD_DN ) {
    player.ty += TARGET_MV;
    if ( player.ty >= SCREEN_HEIGHT-8 ) player.ty = SCREEN_HEIGHT-8;
  }
  if ( (pad & (PAD_A | PAD_B)) ) {
    if ( player.btnPressed == 0 ) {
      player.btnPressed = 1;
      playSe(KICK, 1);
    }
  } else {
    if ( player.btnPressed ) {
      addLaser((player.x>>8), (player.y>>8));
      prvTrgIdx = player.trgIdx;
      player.trgIdx = 0;
      player.btnPressed = 0;
    }
  }
  if ( wdCnt == 15 ) {
    for ( i=0 ; i<9 ; i++ ) {
      if ( i%3 == 0 ) continue;
      plyVt[i].x = randNSW(10<<8, 10<<8);
      plyVt[i].y = randNSW(10<<8, 10<<8);
      plyVt[i].z = randNSW(10<<8, 10<<8);
    }
  } else {
    for ( i=0 ; i<9 ; i++ ) {
      if ( i%3 == 0 ) continue;
      plyVt[i].x -= plyVt[i].x>>4;
      plyVt[i].y -= plyVt[i].y>>4;
      plyVt[i].z -= plyVt[i].z>>4;
    }
  }
  player.cnt++;
  drawObject(&plBmp, (player.x>>8)-4, (player.y>>8)-4, 8, 8);
  drawPrimitive(plyVt, 9, 3, 
		player.x-(SCREEN_WIDTH<<7), player.y-(SCREEN_HEIGHT<<5), (SCREEN_WIDTH<<7));
  if ( player.btnPressed ) {
    drawObject(&trgBmp, player.tx-8, player.ty-8, 16, 16);
  } else {
    drawObject(&nlkBmp, player.tx-8, player.ty-8, 16, 16);
  }
}

#define LASER_COLOR 2

static int drawLaser(Laser *ls, int ax, int ay) {
  int c = ls->cnt, l, x, y, i, sx, ex, sy, ey;
  int nx = 9999, ny = 9999;
  unsigned char *vb;
  l = randN(16) + 4;
  if ( ls->d[c] == 0 ) {
    ny = ls->y[c+1] = ls->y[c];
    ls->d[c+1] = 1;
    x = ls->x[c];
    if ( absN(x-ax) < l ) {
      nx = ls->x[c+1] = ax; 
    } else {
      if ( x > ax ) {
	nx = ls->x[c+1] = x - l;
      } else {
	nx = ls->x[c+1] = x + l;
      }
    }
  } else {
    nx = ls->x[c+1] = ls->x[c];
    ls->d[c+1] = 0;
    y = ls->y[c];
    if ( absN(y-ay) < l ) {
      ny = ls->y[c+1] = ay; 
    } else {
      if ( y > ay ) {
	ny = ls->y[c+1] = y - l;
      } else {
	ny = ls->y[c+1] = y + l;
      }
    }
  }
  ls->cnt++;
  if ( ls->cnt >= LASER_POINT_MAX-1 || absN(nx-ax)+absN(ny-ay) < 4 ) {
    ls->cnt = NOT_EXIST;
    return 1;
  }

  for ( i=1 ; i<=ls->cnt ; i++ ) {
    if ( ls->d[i-1] == 0 ) {
      y = ls->y[i];
      if ( y < 0 || y >= SCREEN_HEIGHT ) continue; 
      sx = ls->x[i-1];
      ex = ls->x[i];
      if ( sx > ex ) {
	int tx = sx;
	sx = ex; ex = tx;
      }
      if ( sx < 0 ) sx = 0;
      if ( ex > SCREEN_WIDTH ) ex = SCREEN_WIDTH;
      l = ex-sx;
      if ( l < 0 ) continue;
      vb = &(vbuf[sx+(y<<7)]);
      memset(vb, LASER_COLOR, l);
    } else {
      x = ls->x[i];
      if ( x < 0 || x >= SCREEN_WIDTH ) continue; 
      sy = ls->y[i-1];
      ey = ls->y[i];
      if ( sy > ey ) {
	int ty = sy;
	sy = ey; ey = ty;
      }
      if ( sy < 0 ) sy = 0;
      if ( ey > SCREEN_HEIGHT ) ey = SCREEN_HEIGHT;
      vb = &(vbuf[x+(sy<<7)]);
      for ( y=sy ; y<ey ; y++, vb+=SCREEN_WIDTH ) {
	*vb = LASER_COLOR;
      }
    }
  }
  return 0;
}
  
void drawLock() {
  int i, x=0, y=0;
  Lock *lck;
  for ( i=0 ; i<LOCK_MAX ; i++ ) {
    if ( lock[i].spc == NOT_EXIST ) continue;
    lck = &(lock[i]);
    if ( lck->spc == TARGET_RE ) {
      if ( lck->re->x == NOT_EXIST ) continue;
      x = lck->re->sx; y = lck->re->sy;
    } else {
      if ( lck->bullet->x == NOT_EXIST ) continue;
      x = lck->bullet->sx; y = lck->bullet->sy;
    }
    drawObject(&numBmp[lck->num+1], x+2, y-10, 8, 8);

    if ( lck->laser.cnt != NOT_EXIST ) {
      if ( drawLaser(&(lck->laser), x, y) ) {
	switch ( lck->num ) {
	case 0:
	case 4:
	  playSe(DOWN1, 2);
	  break;
	case 1:
	case 5:
	  playSe(DOWN2, 2);
	  break;
	case 2:
	case 6:
	  playSe(DOWN3, 2);
	  break;
	case 3:
	case 7:
	  playSe(DOWN4, 2);
	  break;
	}
	if ( lck->spc == TARGET_RE ) {
	  lck->re->x = NOT_EXIST;
	  score += 10 * prvTrgIdx;
	} else {
	  lck->bullet->x = NOT_EXIST;
	  score += prvTrgIdx;
	}
	lck->spc = NOT_EXIST;
      }
    }
  }
}

static void scroolGround() {
  int i, j, ofs;
  for ( i=0 ; i<GROUND_NUM ; i++ ) {
    if ( i == scrGndIdx ) {
      ofs = GROUND_OFFSET*(GROUND_NUM-1);
    } else {
      ofs = -GROUND_OFFSET;
    }
    for ( j=i*4 ; j<i*4+4 ; j++ ) {
      gndVt[j].z += ofs;
    }
  }
  scrGndIdx++; if ( scrGndIdx >= GROUND_NUM ) scrGndIdx = 0;
}

void drawGround() {
  drawPrimitive(gndVt, 4*GROUND_NUM, 4, -(player.x-(SCREEN_WIDTH<<7)), 0, groundZ);
  groundZ -= crtZspd;
  if ( crtZspd > zSpeed ) {
    crtZspd -= (crtZspd>>4);
    if ( crtZspd-zSpeed < 128 ) crtZspd = zSpeed;
  } 
  if ( groundZ < 0 ) {
    scroolGround();
    groundZ += GROUND_OFFSET;
  }
}
